from __future__ import annotations

import dataclasses
import typing as typ

from numbers import Number

import jijmodeling as jm
import numpy as np

from jijzept.entity.schema import SolverType
from jijzept.response import JijModelingResponse
from jijzept.sampler.base_sampler import (
    JijZeptBaseSampler,
    ParameterSearchParameters,
    check_kwargs_against_dataclass,
    merge_params_and_kwargs,
    sample_model,
)
from jijzept.sampler.sample_hubo import sample_hubo
from jijzept.sampler.sample_qubo import sample_qubo
from jijzept.sampler.utils import _set_needs_square

T = typ.TypeVar("T")


class QIOParameters:
    pass


P = typ.TypeVar("P")


class JijQIOBaseSampler(JijZeptBaseSampler, typ.Generic[P]):
    parameter_cls: typ.Type[QIOParameters]

    def sample_model(
        self,
        model: jm.Problem,
        feed_dict: dict[str, Number | list | np.ndarray],
        multipliers: dict[str, Number] = {},
        fixed_variables: dict[str, dict[tuple[int, ...], int | float]] = {},
        needs_square_dict: dict[str, bool] | None = None,
        search: bool = False,
        num_search: int = 15,
        algorithm: str | None = None,
        parameters: P | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Sample using JijModeling by means of Microsoft QIO Simulated Annealing solver.

        To configure the solver, instantiate the `JijQIOXXParameters` class for your desired solver and pass the instance to the `parameters` argument.

        Args:
            model (jijmodeling.Problem): Mathematical model is constracted by JijModeling.
            feed_dict (Dict[str, Union[Number, list, np.ndarray]]): The actual values to be assigned to the placeholders.
            multipliers (Dict[str, Number], optional): The actual multipliers for penalty terms, derived from constraint conditions.
            fixed_variables (Dict[str, Dict[Tuple[int, ...], Union[int, float]]]): dictionary of variables to fix.
            needs_square_dict (Dict[str, bool], optional): If `True`, the corresponding constraint is squared when added as a penalty to the QUBO. When the constraint label is not added to the 'needs_square_dict', it defaults to `True` for linear constraints and `False` for non-linear constraints.
            search (bool, optional): If `True`, the parameter search will be carried out, which tries to find better values of multipliers for penalty terms.
            num_search (int, optional): The number of parameter search iteration. Defaults to set 15. This option works if `search` is `True`.
            algorithm (Optional[str]): Algorithm for parameter search. Defaults to None.
            parameters (P | None, optional): parameters for QIO. Defaults to None.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 60 (one minute) will be set. Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.
            **kwargs: QIO parameters using **kwargs. If both `**kwargs` and `parameters` are exist,　the value of `**kwargs` takes precedence.

        Returns:
            JijModelingResponse: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            import jijmodeling as jm
            n = jm.Placeholder('n')
            x = jm.Binary('x', shape=n)
            d = jm.Placeholder('d', shape=n)
            i = jm.Element("i", n)
            problem = jm.Problem('problem')
            problem += jm.Sum(i, d[i] * x[i])
            sampler = jz.JijQIOSASampler(config='config.toml')
            response = sampler.sample_model(problem, feed_dict={'n': 5, 'd': [1,2,3,4,5]})
            ```
        """
        model = _set_needs_square(model, needs_square_dict)

        check_kwargs_against_dataclass(kwargs, self.parameter_cls)
        param_dict = merge_params_and_kwargs(parameters, kwargs, self.parameter_cls)

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        para_search_params = ParameterSearchParameters(
            multipliers=multipliers,
            mul_search=search,
            num_search=num_search,
            algorithm=algorithm,
        )

        sample_set = sample_model(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            problem=model,
            instance_data=feed_dict,
            fixed_variables=fixed_variables,
            parameter_search_parameters=para_search_params,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set

    def sample_qubo(
        self,
        Q: dict,
        parameters: P | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Microsoft QIO solver is performed for binary quadratic models.

        To configure the solver, instantiate the `JijQIOXXParameters` class for your desired solver and pass the instance to the `parameters` argument.

        Args:
            Q (Dict): The linear or quadratic terms.
            parameters (P | None, optional): parameters for QIO. Defaults to None.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If `None`, 60 (one minute) will be set. Please note that this argument is for the `jijzept` timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.
            **kwargs: QIO parameters using **kwargs. If both `**kwargs` and `parameters` are exist, the value of `**kwargs` takes precedence.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            sampler = jz.JijQIOSASampler(config='config.toml')
            response = sampler.sample_qubo(Q={(0,0): -0.5, (0,1): -1, (0,2): -1, (2,3): -1})
            ```
        """
        check_kwargs_against_dataclass(kwargs, self.parameter_cls)
        param_dict = merge_params_and_kwargs(parameters, kwargs, self.parameter_cls)

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_qubo(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            qubo=Q,
            constant=0,
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )

        return sample_set

    def sample_hubo(
        self,
        J: dict,
        parameters: P | None = None,
        max_wait_time: int | float | None = None,
        sync: bool = True,
        queue_name: str | None = None,
        **kwargs,
    ) -> JijModelingResponse:
        """Microsoft QIO simulated annealing solver is performed for binary polynomial models.

        To configure the solver, instantiate the `JijQIOXXParameters` class for your desired solver and pass the instance to the `parameters` argument.

        Args:
            J (Dict): Polynomial interactions.
            parameters (P | None, optional): parameters for QIO. Defaults to None.
            max_wait_time (int | float | None, optional): The number of timeout [sec] for post request. If None, 60 (one minute) will be set. Please note that this argument is for the jijzept timeout and not for configuring solver settings, such as solving time.
            sync (bool, optional): Synchronous mode.
            queue_name (Optional[str], optional): Queue name.
            **kwargs: QIO parameters using **kwargs. If both `**kwargs` and `parameters` are exist, the value of `**kwargs` takes precedence.

        Returns:
            dimod.SampleSet: Stores minimum energy samples and other information.

        Examples:
            ```python
            import jijzept as jz
            sampler = jz.JijQIOSASampler(config='config.toml')
            response = sampler.sample_hubo(J={(0,1,2,3,4): -1}, vartype="SPIN")
            ```
        """
        check_kwargs_against_dataclass(kwargs, self.parameter_cls)
        param_dict = merge_params_and_kwargs(parameters, kwargs, self.parameter_cls)

        if queue_name is None:
            queue_name = self.jijmodeling_solver_type.queue_name

        sample_set = sample_hubo(
            self.client,
            self.jijmodeling_solver_type.solver,
            queue_name=queue_name,
            polynomial=J,
            constant=J.get((), 0.0),
            max_wait_time=max_wait_time,
            sync=sync,
            **param_dict,
        )
        return sample_set


@dataclasses.dataclass
class JijQIOSAParameters(QIOParameters):
    """Manage Parameters for using Microsoft QIO Simulated Annealing.

    An parameter with none will be set to the default value in QIO Simulated Annealing.
    For more information on the parameters. Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-simulated-annealing#parameterized-simulated-annealing-cpu).

    Attributes:
        num_sweeps (Optional[int], optional): `sweeps` in QIO Simulated Annealing. The number of sweeps. this will be set automatically.
        beta_min (Optional[float], optional): `beta_start` in QIO Simulated Annealing. Minimum (initial) inverse temperature. this will be set automatically.
        beta_max (Optional[float], optional): `beta_stop` in QIO Simulated Annealing. Maximum (final) inverse temperature. this will be set automatically.
        num_reads (Optional[int], optional): `restarts` in QIO Simulated Annealing. The number of samples. this will be set automatically.
        seed (Optional[int], optional): Seed value - used for reproducing results.
        platform (Optional[str]): "CPU" or "FPGA". If `None`, "CPU" will be set.
    """

    num_sweeps: int | None = None
    beta_min: float | None = None
    beta_max: float | None = None
    num_reads: int | None = None
    seed: int | None = None
    platform: str | None = None


class JijQIOSASampler(JijQIOBaseSampler[JijQIOSAParameters]):
    """Sampler for using Microsoft QIO Simulated Annealing.

    For details on Microsoft QIO Simulated Annealing, Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-simulated-annealing).

    """

    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="SA")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="SAParaSearch")
    parameter_cls: typ.Type[QIOParameters] = JijQIOSAParameters


@dataclasses.dataclass
class JijQIOPAParameters(QIOParameters):
    """Manage Parameters for using Microsoft QIO Population Annealing.

    An parameter with none will be set to the default value in QIO Population Annealing.
    For more information on the parameters. Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-population-annealing#parameterized-population-annealing).

    Attributes:
        num_sweeps (Optional[int], optional): `sweeps` in QIO Population Annealing. The number of sweeps. this will be set automatically.
        beta_min (Optional[float], optional): Initial value of annealing schedule `beta` in QIO Population Annealing. this will be set automatically.
        beta_max (Optional[float], optional): Final value of annealing schedule `beta` in QIO Population Annealing.  this will be set automatically.
        population (Optional[int], optional): The Number of walkers in the population. this will be set automatically.
        schedule_type (Optional[str], optional): Type of annealing schedule `beta` in QIO Population Annealing. Specifying 'linear' or 'geometric'. If `None`, this will be set 'linear'.
        seed (Optional[int], optional): Seed value - used for reproducing results.
    """

    num_sweeps: int | None = None
    beta_min: float | None = None
    beta_max: float | None = None
    population: int | None = None
    schedule_type: str | None = None
    seed: int | None = None


class JijQIOPASampler(JijQIOBaseSampler[JijQIOPAParameters]):
    """Smapler for using Microsoft QIO Population Annealing.

    For details on Microsoft QIO Population Annealing, Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-population-annealing).

    """

    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="PA")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="PAParaSearch")
    parameter_cls = JijQIOPAParameters


@dataclasses.dataclass
class JijQIOPTParameters(QIOParameters):
    """Manage Parameters for using Microsoft QIO Parallel Tempering.

    An parameter with none will be set to the default value in QIO Parallel Tempering.
    For more information on the parameters. Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-parallel-tempering#parameterized-parallel-tempering).

    Attributes:
        num_sweeps (Optional[int], optional): `sweeps` in QIO Parallel Tempering. The number of sweeps. this will be set automatically.
        replicas (Optional[int], optional): The number of concurrent running copies for sampling. `replicas` must be the length of `all_betas`.
        all_betas (Union[int, float, None], optional): The list of beta values used in each replica for sampling. this will be set automatically.
        seed (Optional[int], optional): Seed value - used for reproducing results.
    """

    num_sweeps: int | None = None
    replicas: int | None = None
    all_betas: list[int | float] | None = None
    seed: int | None = None


class JijQIOPTSampler(JijQIOBaseSampler[JijQIOPTParameters]):
    """Sampler for using Microsoft QIO Parallel tempering.

    For details on Microsoft QIO Parallel tempering, Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-parallel-tempering).

    """

    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="PT")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="PTParaSearch")
    parameter_cls: typ.Type[QIOParameters] = JijQIOPTParameters


@dataclasses.dataclass
class JijQIOTBParameters(QIOParameters):
    """Manage Parameters for using Microsoft QIO Tabu Search.

    An parameter with none will be set to the default value in QIO Tabu Search.
    For more information on the parameters. Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-tabu-search#parameterized-tabu-search).

    Attributes:
        num_sweeps (Optional[int], optional): `sweeps` in QIO Tabu Search. The number of sweeps. this will be set automatically.
        tabu_tenure (Optional[int], optional): Tenure of the tabu list in moves.  this will be set automatically.
        num_reads (Optional[int], optional): `restarts` in QIO Tabu Search. The number of samples. this will be set automatically.
        seed (Optional[int], optional): Seed value - used for reproducing results.
    """

    num_sweeps: int | None = None
    tabu_tenure: int | None = None
    num_reads: int | None = None
    seed: int | None = None


class JijQIOTBSampler(JijQIOBaseSampler):
    """Sampler for using Microsoft Tabu Search.

    For details on Microsoft QIO Tabu Search, Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-tabu-search).

    """

    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="TB")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="TBParaSearch")
    parameter_cls: typ.Type[QIOParameters] = JijQIOTBParameters


@dataclasses.dataclass
class JijQIOSQAParameters(QIOParameters):
    """Manage Parameters for using Microsoft QIO Quantum Monte Carlo.

    An parameter with none will be set to the default value in QIO Quantum Monte Carlo.
    For more information on the parameters. Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-quantum-monte-carlo#parameterized-quantum-monte-carlo).

    Attributes:
        num_sweeps (Optional[int], optional): `sweeps` in QIO Quantum Monte Carlo. The number of sweeps. this will be set automatically.
        num_reads (Optional[int], optional): `restarts` in QIO Quantum Monte Carlo. The number of samples. this will be set automatically.
        beta (Optional[float], optional): `beta_start` in QIO Quantum Monte Carlo. Minimum (initial) inverse temperature. this will be set automatically.
        trotter (Optional[int], optional): `trotter_number` in QIO Quantum Monte Carlo. The number of Trotter. this will be set automatically.
        transverse_field_max (Optional[float], optional): `transverse_field_start` in QIO Quantum Monte Carlo. The starting values of the external field applied to the annealing schedule. this will be set automatically.
        transverse_field_min (Optional[float], optional): `transverse_field_stop` in "QIO Quantum Monte Carlo. The stopping values of the external field applied to the annealing schedule. this will be set automatically.
        seed (Optional[int], optional): Seed value - used for reproducing results.
    """

    num_sweeps: int | None = None
    num_reads: int | None = None
    beta: float | None = None
    trotter: int | None = None
    transverse_field_max: float | None = None
    transverse_field_min: float | None = None
    seed: int | None = None


class JijQIOSQASampler(JijQIOBaseSampler):
    """Sampler for using Microsoft QIO Quantum Monte Carlo.

    For details on Microsoft QIO Quantum Monte Carlo, Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-quantum-monte-carlo).

    """

    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="SQA")
    jijmodeling_solver_type = SolverType(queue_name="qiosolver", solver="SQAParaSearch")
    parameter_cls: typ.Type[QIOParameters] = JijQIOSQAParameters


@dataclasses.dataclass
class JijQIOSSMCParameters(QIOParameters):
    """Manage Parameters for using Microsoft QIO Substochastic Monte Carlo.

    An parameter with none will be set to the default value in QIO Substochastic Monte Carlo.
    For more information on the parameters. Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-substochastic-monte-carlo#parameterized-substochastic-monte-carlo).

    Attributes:
        search (bool, optional): If `True`, the parameter search will be carried out, which tries to find bettervalues of multipliers for penalty terms.
        num_search (int, optional): The number of parameter search iteration. This option works if `search` is `True`.
        step_limit (Optional[int], optional): The number of Monte Carlo steps. this will be set automatically.
        target_population (Optional[int], optional): The number of walkers in the population. this will be set automatically.
        alpha_min (Optional[float], optional): Initial value of `alpha` in QIOSSMC. `alpha` is the schedule for the stepping chance. this will be set automatically.
        alpha_max (Optional[float], optional): Final value of `alpha` in QIOSSMC. `alpha` is the schedule for the stepping chance. this will be set automatically.
        beta_min (Optional[float], optional): Initial value of `beta` in QIOSSMC. `beta` is the schedule for the resampling factor. this will be set automatically.
        beta_max (Optional[float], optional): Final value of `beta` in QIOSSMC. `beta` is the schedule for the resampling factor. this will be set automatically.
        alpha_schedule_type (Optional[str], optional): Schedule type of `alpha` in QIOSSMC. 'linear' or 'geometric'. If `None` will be set 'linear'.
        beta_schedule_type (Optional[str], optional): Schedule type of `beta` in QIOSSMC. 'linear' or 'geometric'. If `None`, this will be set 'linear'.
        seed (Optional[int], optional): Seed value - used for reproducing results.
    """

    search: bool = False
    num_search: int = 15
    algorithm: str | None = None
    step_limit: int | None = None
    target_population: int | None = None
    alpha_min: float | None = None
    alpha_max: float | None = None
    beta_min: float | None = None
    beta_max: float | None = None
    alpha_schedule_type: str | None = None
    beta_schedule_type: str | None = None
    seed: int | None = None


class JijQIOSSMCSampler(JijQIOBaseSampler[JijQIOSSMCParameters]):
    """Sampler for using Substochastic Monte Carlo.

    For details on Microsoft QIO Substochastic Monte Carlo, Please see [this page](https://learn.microsoft.com/en-us/azure/quantum/optimization-substochastic-monte-carlo).

    """

    solver_type: SolverType = SolverType(queue_name="qiosolver", solver="SSMC")
    jijmodeling_solver_type = SolverType(
        queue_name="qiosolver", solver="SSMCParaSearch"
    )
    parameter_cls: typ.Type[QIOParameters] = JijQIOSSMCParameters
