from __future__ import annotations

import itertools

import jijmodeling as jm
import numpy as np

from jijzept.client import JijZeptClient
from jijzept.sampler.base_sampler import ParameterSearchParameters, sample_model


def hubo_problem(
    hubo: dict[tuple[int, ...], float] | dict[tuple[str, ...], float], constant: float
) -> tuple[
    jm.Problem, dict[str, np.ndarray | list], dict[int, str], dict[str, jm.Binary]
]:
    """Convert the coefficients of a HUBO represented as a dictionary to a JijModeling Problem object.

    Args:
    - hubo: A dictionary representing a HUBO problem, where the keys are tuples of integers or strings
           representing the indices of the decision variables, and the values are floats representing
           the coefficients of the high-order terms of the HUBO.
    - constant: A float representing the constant energy offset in the HUBO problem.

    Returns:
    - A tuple containing:
        - The HUBO problem as a JijModeling Problem object.
        - The instance data as a dictionary containing the values of the placeholders of the problem.
        - A dictionary mapping the indices of the decision variables to their corresponding labels (if the
          indices are represented as strings).
        - A dictionary containing the decision variables as JijModeling Binary objects.
    """
    hubo_keys = list(hubo.keys())
    unique_labels: dict[str, int] = {}
    N = jm.Placeholder("N")
    key = hubo_keys[0][0]
    if isinstance(key, int):
        # decision_variables = [jm.Binary("x", shape=N)]
        pass
    elif isinstance(key, str):
        unique_labels = {
            k: i for i, k in enumerate(itertools.chain.from_iterable(hubo_keys))
        }
        hubo_keys = [[unique_labels[e0], unique_labels[e1]] for e0, e1 in hubo_keys]
    else:
        TypeError(
            f"`{key.__class__.__name__}` object is not supported as a key of HUBO."
        )

    N = jm.Placeholder("N")  # max variable index
    E = jm.JaggedArray("E", dim=2)  # super graph edge
    L = E.shape[0]  # number of edges
    Q = jm.Placeholder("Q", shape=L)  # weight of edge
    c = jm.Placeholder("c")  # constant
    e = jm.Element("e", (0, L))  # edge index

    d = E[e].shape[0]  # order of edge 'e'
    k = jm.Element("k", (0, d))

    x = jm.Binary("x", shape=N)  # decision variable

    hubo_model = jm.Problem("hubo")
    hubo_model += jm.Sum(e, Q[e] * jm.Prod(k, x[E[e, k]])) + c

    edges = [list(k) for k in hubo.keys()]
    values = np.array(list(hubo.values()), dtype=float)
    instance_data = {
        "N": len(edges),
        "E": edges,
        "Q": values.tolist(),
        "c": constant,
    }

    return hubo_model, instance_data, {i: k for k, i in unique_labels.items()}, {"x": x}


def sample_hubo(
    client: JijZeptClient,
    solver: str,
    queue_name: str,
    polynomial: dict[tuple[int, ...], float],
    constant: float,
    max_wait_time: int | float | None,
    sync: bool,
    **kwargs,
):
    """Sample HUBO problem using the specified solver.

    Args:
    - client: A JijZeptClient object representing the client used to submit the problem.
    - solver: A string representing the name of the solver used to solve the problem.
    - queue_name: A string representing the name of the queue to use for the problem.
    - polynomial: A dictionary representing a HUBO problem, where the keys are tuples of integers representing
           the indices of the decision variables, and the values are floats representing the coefficients
           of the high-order terms of the HUBO.
    - constant: A float representing the constant energy offset in the HUBO problem.
    - max_wait_time: An integer or float representing the maximum time to wait for the problem to complete
                     (in seconds). If None, wait indefinitely.
    - sync: A boolean representing whether to wait for the problem to complete before returning the results.
    - **kwargs: Additional arguments to pass to the sample_model function.

    Returns:
    - A SampleSet object: Stores minimum energy samples and other information.
    """
    hubo_model, instance_data, var_map, _ = hubo_problem(polynomial, constant)

    parameters = ParameterSearchParameters(
        multipliers={}, mul_search=False, normalize_qubo=False
    )

    sample_set = sample_model(
        client=client,
        solver=solver,
        queue_name=queue_name,
        problem=hubo_model,
        instance_data=instance_data,
        fixed_variables={},
        parameter_search_parameters=parameters,
        max_wait_time=max_wait_time,
        sync=sync,
        **kwargs,
    )
    sample_set.set_variable_labels(var_map)

    return sample_set
